__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import logging
from itertools import combinations
from pathlib import Path
from time import time
from typing import Dict, List, Iterator, Optional

from imagehash import ImageHash

from duplicate_images.function_types import Cache, HashFunction, ImagePair, Results
from duplicate_images.hash_scanner import ImageHashScanner, ParallelImageHashScanner
from duplicate_images.pair_finder_options import PairFinderOptions


class ImagePairFinder(ImageHashScanner):

    @classmethod
    def create(
            cls, files: List[Path], hash_algorithm: HashFunction,
            options: PairFinderOptions = PairFinderOptions(),
            hash_store: Optional[Cache] = None
    ) -> 'ImagePairFinder':
        if options.max_distance == 0 and not options.slow:
            if not options.parallel:
                return DictImagePairFinder(
                    files, hash_algorithm, options=options, hash_store=hash_store
                )
            return ParallelDictImagePairFinder(
                files, hash_algorithm, options=options, hash_store=hash_store
            )
        if options.parallel:
            return ParallelSlowImagePairFinder(
                files, hash_algorithm, options=options, hash_store=hash_store
            )
        return SlowImagePairFinder(files, hash_algorithm, options=options, hash_store=hash_store)

    def __init__(  # pylint: disable = too-many-arguments
            self, files: List[Path], hash_algorithm: HashFunction,
            options: PairFinderOptions = PairFinderOptions(),
            hash_store: Optional[Cache] = None
    ) -> None:
        logging.info('Using %s', self.__class__.__name__)
        self.precalculated_hashes: Dict = {}
        super().__init__(files, hash_algorithm, options, hash_store)
        self.files = files
        self.scan_start_time = time()

    def get_pairs(self) -> Results:
        raise NotImplementedError()

    def log_scan_finished(self) -> None:
        logging.info(
            '%d hashes calculated in %.2fs',
            len(self.precalculated_hashes), time() - self.scan_start_time
        )


class SlowImagePairFinder(ImagePairFinder, ImageHashScanner):
    """
    Searches by comparing the image hashes of each image to every other, giving O(N^2) performance.
    The only option if max_distance != 0.
    """

    def __init__(  # pylint: disable = too-many-arguments
            self, files: List[Path], hash_algorithm: HashFunction,
            options: PairFinderOptions = PairFinderOptions(),
            hash_store: Optional[Cache] = None
    ) -> None:
        super().__init__(files, hash_algorithm, options, hash_store)
        self.precalculated_hashes = self.get_hashes(files)
        self.progress_bars.close_reader()

    def get_pairs(self) -> Results:
        self.log_scan_finished()
        image_files = list(self.precalculated_hashes.keys())
        all_pairs = (
            (file, other_file)
            for file in image_files
            for other_file in image_files[image_files.index(file) + 1:]
        )
        logging.info('Filtering duplicates')
        matches = self.filter_matches(all_pairs)
        self.progress_bars.close()
        return matches

    def get_hashes(self, image_files: List[Path]) -> Dict[Path, ImageHash]:
        return {
            file: image_hash for file, image_hash in self.precalculate_hashes(image_files)
            if image_hash is not None
        }

    def filter_matches(self, all_pairs: Iterator[ImagePair]) -> Results:
        self.progress_bars.create_filter_bar(len(self.precalculated_hashes))
        return [
            (file, other_file) for file, other_file in all_pairs
            if self.are_images_equal(file, other_file)
        ]

    def are_images_equal(self, file: Path, other_file: Path) -> bool:
        self.progress_bars.update_filter()
        hash_distance = self.precalculated_hashes[file] - self.precalculated_hashes[other_file]
        logging.debug(
            '%-30s - %-30s = %d', file.stem, other_file.stem, hash_distance
        )
        return hash_distance <= self.max_distance


class DictImagePairFinder(ImagePairFinder, ImageHashScanner):
    """
    Searches by storing the image hashes as keys to a dict.
    Works only if max_distance == 0.
    """
    def __init__(  # pylint: disable = too-many-arguments
            self, files: List[Path], hash_algorithm: HashFunction,
            options: PairFinderOptions = PairFinderOptions(),
            hash_store: Optional[Cache] = None
    ) -> None:
        super().__init__(files, hash_algorithm, options, hash_store)
        if options.max_distance != 0:
            raise ValueError('DictImagePairFinder only works if max_distance == 0!')
        self.precalculated_hashes = self.get_hashes(files)
        self.progress_bars.close_reader()

    def get_pairs(self) -> Results:
        self.progress_bars.close()
        self.log_scan_finished()
        results = [
            pair
            for result in self.precalculated_hashes.values()
            for pair in combinations(list(result), 2)
            if len(result) > 1
        ]
        return results

    def get_hashes(self, image_files: List[Path]) -> Dict[ImageHash, List[Path]]:
        hash_dict: Dict[ImageHash, List[Path]] = {}
        for file, image_hash in self.precalculate_hashes(image_files):
            if image_hash is not None:
                hash_dict.setdefault(image_hash, []).append(file)
        return hash_dict


class ParallelDictImagePairFinder(DictImagePairFinder, ParallelImageHashScanner):
    pass


class ParallelSlowImagePairFinder(SlowImagePairFinder, ParallelImageHashScanner):
    pass



#!/usr/bin/env /usr/bin/python3

import logging
from argparse import Namespace
from os import walk, access, R_OK
from pathlib import Path
from typing import Callable, List, Optional

from filetype import guess
from pillow_heif import register_heif_opener

from duplicate_images.common import path_with_parent
from duplicate_images.function_types import Results
from duplicate_images.hash_store import PickleHashStore
from duplicate_images.image_pair_finder import ImagePairFinder, PairFinderOptions
from duplicate_images.log import setup_logging
from duplicate_images.methods import ACTIONS_ON_EQUALITY, IMAGE_HASH_ALGORITHM
from duplicate_images.parse_commandline import parse_command_line

register_heif_opener()


def is_image_file(filename: Path) -> bool:
    """Returns True if filename is a readable image file"""
    if access(filename, R_OK) and not filename.is_symlink():
        kind = guess(filename)
        return kind is not None and kind.mime.startswith('image/')
    return False


def files_in_dirs(
        dir_names: List[Path], is_file: Callable[[Path], bool] = lambda f: f.is_file()
) -> List[Path]:
    """Returns a list of all files in directory dir_name, recursively scanning subdirectories"""
    files = [
        Path(root) / filename
        for dir_name in dir_names
        for root, _, filenames in walk(dir_name)
        for filename in filenames
        if is_file(Path(root) / filename)
    ]
    return files


def get_matches(
        root_directories: List[Path], algorithm: str,
        options: PairFinderOptions = PairFinderOptions(),
        hash_store_path: Optional[Path] = None
) -> Results:
    hash_algorithm = IMAGE_HASH_ALGORITHM[algorithm]
    image_files = sorted(files_in_dirs(root_directories, is_image_file))
    logging.info('%d total files', len(image_files))
    logging.info('Computing image hashes')

    with PickleHashStore.create(hash_store_path) as hash_store:
        return ImagePairFinder.create(
            image_files, hash_algorithm, options=options, hash_store=hash_store
        ).get_pairs()


def execute_actions(matches: Results, args: Namespace) -> None:
    action_equal = ACTIONS_ON_EQUALITY[args.on_equal]

    for pair in sorted(matches):
        try:
            action_equal(args, pair)
        except FileNotFoundError:
            continue


def main() -> None:
    args = parse_command_line()
    setup_logging(args)
    options = PairFinderOptions.from_args(args)
    for folder in args.root_directory:
        logging.info('Scanning %s', path_with_parent(folder))
    try:
        matches = get_matches(
            [Path(folder) for folder in args.root_directory], args.algorithm,
            options=options, hash_store_path=Path(args.hash_db) if args.hash_db else None
        )
        logging.info('%d matches', len(matches))
        execute_actions(matches, args)
    except KeyboardInterrupt:
        pass


if __name__ == '__main__':
    main()



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

from argparse import Namespace
from dataclasses import dataclass
from typing import Optional


@dataclass(frozen=True)
class PairFinderOptions:
    max_distance: int = 0
    hash_size: Optional[int] = None
    show_progress_bars: bool = False
    parallel: bool = False
    slow: bool = False

    @classmethod
    def from_args(cls, args: Namespace):
        return cls(
            args.max_distance, args.hash_size, args.progress, args.parallel, args.slow
        )



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

import json
import logging
import pickle  # nosec
from pathlib import Path
from typing import Any, IO, Callable, Optional, Union

from imagehash import ImageHash, hex_to_hash

from duplicate_images.function_types import Cache


class NullHashStore:

    def __init__(self) -> None:
        logging.info('No persistent storage for calculated image hashes set up')

    def __enter__(self) -> Cache:
        return {}

    def __exit__(self, _: Any, __: Any, ___: Any) -> None:
        pass


class FileHashStore:

    @classmethod
    def create(cls, store_path: Optional[Path]) -> Union['FileHashStore', NullHashStore]:
        if store_path is None:
            return NullHashStore()
        if store_path.suffix == '.pickle':
            return PickleHashStore(store_path)
        return JSONHashStore(store_path)

    def __init__(self, store_path: Path) -> None:
        self.store_path = store_path
        self.values: Cache = {}
        try:
            self.load()
            logging.info(
                'Opened persistent storage %s with %d entries', store_path, len(self.values)
            )
        except (FileNotFoundError, EOFError, pickle.PickleError):
            logging.info('Creating new persistent storage at %s', store_path)

    def __enter__(self) -> Cache:
        return self.values

    def __exit__(self, _: Any, __: Any, ___: Any) -> None:
        if self.store_path.is_file():
            if self.store_path.with_suffix('.bak').is_file():
                self.store_path.with_suffix('.bak').unlink()
            self.store_path.rename(self.store_path.with_suffix('.bak'))
        self.dump()

    def load(self) -> None:
        raise NotImplementedError()

    def dump(self) -> None:
        raise NotImplementedError()


class PickleHashStore(FileHashStore):

    def load(self) -> None:
        with self.store_path.open('rb') as file:
            self.values = checked_load(file, pickle.load)

    def dump(self) -> None:
        with self.store_path.open('wb') as file:
            pickle.dump(self.values, file)  # nosec


class JSONHashStore(FileHashStore):

    def load(self) -> None:
        with self.store_path.open('r') as file:
            self.values = checked_load(
                file,
                lambda f: {Path(k): hex_to_hash(v) for k, v in json.load(f).items()}
            )

    def dump(self) -> None:
        with self.store_path.open('w') as file:
            json.dump({str(k.resolve()): str(v) for k, v in self.values.items()}, file)


def checked_load(file: IO, load: Callable[[IO], Cache]) -> Cache:
    values = load(file)  # nosec
    if not isinstance(values, dict):
        raise ValueError(f'Not a dict: {values}')
    bad_keys = [key for key in values.keys() if not isinstance(key, Path)]
    if bad_keys:
        raise ValueError(f'Not a Path: {bad_keys}')
    bad_values = [value for value in values.values() if not isinstance(value, ImageHash)]
    if bad_values:
        raise ValueError(f'Not an image hash: {bad_values}')
    return values


HashStore = Union[NullHashStore, PickleHashStore, JSONHashStore]
